import inspect
from inspect import isasyncgenfunction, iscoroutinefunction
from typing import Callable, Generic, List, Optional, Type, TypeVar

from cached_property import cached_property  # type: ignore

from strawberry.arguments import get_arguments_from_annotations
from strawberry.exceptions import MissingArgumentsAnnotationsError
from strawberry.types.types import ArgumentDefinition
from strawberry.utils.inspect import get_func_args


T = TypeVar("T")


class StrawberryResolver(Generic[T]):
    def __init__(self, func: Callable[..., T], *, description: Optional[str] = None):
        self.wrapped_func = func
        self._description = description

    # TODO: Use this when doing the actual resolving? How to deal with async resolvers?
    def __call__(self, *args, **kwargs) -> T:
        return self.wrapped_func(*args, **kwargs)

    # TODO: Return StrawberryArguments instead. Maybe this should be a classmethod
    #       there instead?
    # TODO: Return a set instead
    @cached_property
    def arguments(self) -> List[ArgumentDefinition]:
        # TODO: Move to StrawberryArgument? StrawberryResolver ClassVar?
        SPECIAL_ARGS = {"root", "self", "info"}

        annotations = self.wrapped_func.__annotations__
        parameters = inspect.signature(self.wrapped_func).parameters
        function_arguments = set(parameters) - SPECIAL_ARGS

        annotations = {
            name: annotation
            for name, annotation in annotations.items()
            if name not in (SPECIAL_ARGS | {"return"})
        }

        annotated_arguments = set(annotations)
        arguments_missing_annotations = function_arguments - annotated_arguments

        if any(arguments_missing_annotations):
            raise MissingArgumentsAnnotationsError(
                field_name=self.wrapped_func.__name__,
                arguments=arguments_missing_annotations,
            )

        return get_arguments_from_annotations(
            annotations, parameters, origin=self.wrapped_func
        )

    @cached_property
    def has_info_arg(self) -> bool:
        args = get_func_args(self.wrapped_func)
        return "info" in args

    @cached_property
    def has_root_arg(self) -> bool:
        args = get_func_args(self.wrapped_func)
        return "root" in args

    @cached_property
    def has_self_arg(self) -> bool:
        args = get_func_args(self.wrapped_func)
        return args and args[0] == "self"

    @cached_property
    def name(self) -> str:
        # TODO: What to do if resolver is a lambda?
        return self.wrapped_func.__name__

    @cached_property
    def type(self) -> Type[T]:
        return self.wrapped_func.__annotations__.get("return", None)

    @cached_property
    def is_async(self) -> bool:
        return iscoroutinefunction(self.wrapped_func) or isasyncgenfunction(
            self.wrapped_func
        )


__all__ = ["StrawberryResolver"]



import dataclasses
from typing import Any, Dict, List, Optional

from graphql.error.graphql_error import GraphQLError


@dataclasses.dataclass
class ExecutionContext:
    query: str
    context: Any = None
    variables: Optional[Dict[str, Any]] = None
    operation_name: Optional[str] = None


@dataclasses.dataclass
class ExecutionResult:
    data: Optional[Dict[str, Any]]
    errors: Optional[List[GraphQLError]]
    extensions: Optional[Dict[str, Any]] = None



import dataclasses
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Type, Union

from strawberry.permission import BasePermission
from strawberry.union import StrawberryUnion


if TYPE_CHECKING:
    from strawberry.types.fields.resolver import StrawberryResolver

undefined = object()


@dataclasses.dataclass
class FederationTypeParams:
    keys: List[str] = dataclasses.field(default_factory=list)
    extend: bool = False


@dataclasses.dataclass
class TypeDefinition:
    name: str
    is_input: bool
    is_interface: bool
    is_generic: bool
    origin: Type
    description: Optional[str]
    federation: FederationTypeParams
    interfaces: List["TypeDefinition"]

    _fields: List["FieldDefinition"]
    _type_params: Dict[str, Type] = dataclasses.field(default_factory=dict, init=False)

    def get_field(self, name: str) -> Optional["FieldDefinition"]:
        return next((field for field in self.fields if field.name == name), None)

    @property
    def fields(self) -> List["FieldDefinition"]:
        from .type_resolver import _resolve_types

        return _resolve_types(self._fields)

    @property
    def type_params(self) -> Dict[str, Type]:
        if not self._type_params:
            from .type_resolver import _get_type_params

            self._type_params = _get_type_params(self.fields)

        return self._type_params


@dataclasses.dataclass
class ArgumentDefinition:
    name: Optional[str] = None
    origin_name: Optional[str] = None
    type: Optional[Type] = None
    origin: Optional[Type] = None
    child: Optional["ArgumentDefinition"] = None
    is_subscription: bool = False
    is_optional: bool = False
    is_child_optional: bool = False
    is_list: bool = False
    is_union: bool = False
    description: Optional[str] = None
    default_value: Any = undefined


@dataclasses.dataclass
class FederationFieldParams:
    provides: List[str] = dataclasses.field(default_factory=list)
    requires: List[str] = dataclasses.field(default_factory=list)
    external: bool = False


@dataclasses.dataclass
class FieldDefinition:
    name: Optional[str]
    origin_name: Optional[str]
    type: Optional[Union[Type, StrawberryUnion]]
    origin: Optional[Union[Type, Callable]] = None
    child: Optional["FieldDefinition"] = None
    is_subscription: bool = False
    is_optional: bool = False
    is_child_optional: bool = False
    is_list: bool = False
    is_union: bool = False
    federation: FederationFieldParams = dataclasses.field(
        default_factory=FederationFieldParams
    )
    arguments: List[ArgumentDefinition] = dataclasses.field(default_factory=list)
    description: Optional[str] = None
    base_resolver: Optional["StrawberryResolver"] = None
    permission_classes: List[Type[BasePermission]] = dataclasses.field(
        default_factory=list
    )
    default_value: Any = undefined
    deprecation_reason: Optional[str] = None



import enum
import inspect
from typing import Any, Dict, List, Mapping, Optional, Type, cast

from typing_extensions import Annotated, get_args, get_origin

from .exceptions import MultipleStrawberryArgumentsError, UnsupportedTypeError
from .scalars import is_scalar
from .types.type_resolver import resolve_type
from .types.types import ArgumentDefinition, undefined
from .utils.str_converters import to_camel_case


class StrawberryArgument:
    description: Optional[str]

    def __init__(self, description: Optional[str] = None):
        self.description = description


def get_arguments_from_annotations(
    annotations: Any, parameters: Mapping[str, inspect.Parameter], origin: Any
) -> List[ArgumentDefinition]:
    arguments = []

    for name, annotation in annotations.items():
        default_value = parameters[name].default
        default_value = (
            undefined
            if default_value is inspect.Parameter.empty or is_unset(default_value)
            else default_value
        )

        argument_definition = ArgumentDefinition(
            origin_name=name,
            name=to_camel_case(name),
            origin=origin,
            default_value=default_value,
        )

        if get_origin(annotation) is Annotated:
            annotated_args = get_args(annotation)

            # The first argument to Annotated is always the underlying type
            argument_definition.type = annotated_args[0]

            argument_metadata = None
            # Find any instances of StrawberryArgument in the other Annotated args,
            # raising an exception if there are multiple StrawberryArguments
            for arg in annotated_args[1:]:
                if isinstance(arg, StrawberryArgument):
                    if argument_metadata is not None:
                        raise MultipleStrawberryArgumentsError(
                            field_name=origin.__name__, argument_name=name
                        )
                    argument_metadata = arg

            if argument_metadata is not None:
                argument_definition.description = argument_metadata.description
        else:
            argument_definition.type = annotation

        arguments.append(argument_definition)

        resolve_type(argument_definition)

    return arguments


class _Unset:
    def __str__(self):
        return ""

    def __bool__(self):
        return False


UNSET: Any = _Unset()


def is_unset(value: Any) -> bool:
    return type(value) is _Unset


def convert_argument(value: Any, argument_definition: ArgumentDefinition) -> Any:
    if value is None:
        return None

    if is_unset(value):
        return value

    if argument_definition.is_list:
        child_definition = cast(ArgumentDefinition, argument_definition.child)

        return [convert_argument(x, child_definition) for x in value]

    argument_type = cast(Type, argument_definition.type)

    if is_scalar(argument_type):
        return value

    # Convert Enum fields to instances using the value. This is safe
    # because graphql-core has already validated the input.
    if isinstance(argument_type, enum.EnumMeta):
        return argument_type(value)  # type: ignore

    if hasattr(argument_type, "_type_definition"):
        assert argument_type._type_definition.is_input

        kwargs = {}

        for field in argument_type._type_definition.fields:
            if field.name in value:
                kwargs[field.origin_name] = convert_argument(value[field.name], field)

        return argument_type(**kwargs)

    raise UnsupportedTypeError(argument_type)


def convert_arguments(
    value: Dict[str, Any],
    arguments: List[ArgumentDefinition],
) -> Dict[str, Any]:
    """Converts a nested dictionary to a dictionary of actual types.

    It deals with conversion of input types to proper dataclasses and
    also uses a sentinel value for unset values."""

    if not arguments:
        return {}

    kwargs = {}

    for argument in arguments:
        if argument.name in value:
            origin_name = cast(str, argument.origin_name)
            current_value = value[argument.name]

            kwargs[origin_name] = convert_argument(current_value, argument)

    return kwargs


def argument(description: Optional[str] = None) -> StrawberryArgument:
    return StrawberryArgument(description=description)


__all__ = [
    "ArgumentDefinition",
    "StrawberryArgument",
    "UNSET",
    "argument",
    "convert_argument",
    "convert_arguments",
    "get_arguments_from_annotations",
    "is_unset",
    "undefined",
]
